# Copyright (c) 2023 Sylk.build.

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import time
import logging
from sylk.builder.src import lru
from sylk.commons import pretty
from sylk.commons import helpers
from sylk.commons.file_system import rFile, wFile, join_path, walkFiles, mkdir
from sylk.commons.helpers import SylkJsonToMessage, MessageToDict
from sylk.architect.commands import GetSylkJson, SaveSylkJson
from sylk.architect.recievers import Core
from sylk.architect.interfaces import IUndoRedo

log = logging.getLogger("sylk.cli.main")

try:
    import cPickle as pickle
except:
    import pickle


class CoreInvoker:
    """The Invoker Class"""

    def __init__(self, path):
        log.debug("Invoker class __init__")
        self._commands = {}
        self._history = []
        self._path = path
        self._sylk_json = {}

    @property
    def history(self):
        """Return all records in the History list"""
        return self._history

    def register(self, command_name, command):
        """All commands are registered in the Invoker Class"""
        self._commands[command_name] = command

    def execute(self, command_name, *args):
        """Execute a pre defined command and log in history"""
        if command_name in self._commands.keys():
            response = self._commands[command_name].execute(self._path, args)
            self._history.append((time.time(), command_name, args, response))
        else:
            log.error(f"Command [{command_name}] not recognised")


class Sylk(IUndoRedo):
    def __init__(self, path=None, load=None, format="json"):
        log.debug("Architect class __init__ | path to project -> {0}".format(path))
        self._saves = 0
        self._format = format
        self._load = load
        self._core = Core()
        self._core_invoker = CoreInvoker(path)
        self.init_core()
        # self._cache = lru.LruCache(100)
        self._commands = {}
        self._hooks = {}
        self._sylk_json = {}
        self._history_position = 0
        self._get_sylk_json()
        self._get_saves()
        temp_sylk = self._sylk_json.copy()
        self._history = [(temp_sylk, "INIT", (path,))]
        self._unsaved_changes = False

    def init_core(self):
        self._core_invoker.register("GetSylkJson", GetSylkJson(self._core))
        self._core_invoker.register("SaveSylkJson", SaveSylkJson(self._core))

    def _get_sylk_json(self):
        self._core_invoker.execute("GetSylkJson",{'format': self.sylkJson.get('configs',{"format":self._format}).get('format','json')})
        self._sylk_json = self._core_invoker.history[-1][-1]

    def _get_saves(self):
        if self._sylk_json.get("project").get("uri") is not None:
            cache = walkFiles(
                join_path(self._sylk_json.get("project").get("uri"), ".sylk", "cache")
            )

            if cache is not None:
                self._saves = len(cache)
                if self._load is not None:
                    loaded_save = next((f for f in cache if f == self._load), None)
                    if loaded_save is None:
                        log.error(f"Couldnt load cached save [{self._load}]")
                    else:
                        path = join_path(
                            self._sylk_json.get("project").get("uri"),
                            ".sylk",
                            "cache",
                            f"{loaded_save}",
                        )
                        sylkJson = SylkJsonToMessage(rFile(path, json=True))
                        # log.error(sylkJson)
                        self._sylk_json = MessageToDict(sylkJson)

    def registerCommand(self, command_name, command):
        """All commands are registered in the Invoker Class"""
        self._commands[command_name] = command

    def registerHook(self, command_name, hook_name, hook, type="before"):
        """All hooks are registered in the Invoker Class"""
        if self._hooks.get(command_name) is None:
            self._hooks[command_name] = {"before": {}, "after": {}}

        self._hooks[command_name][type][hook_name] = hook

    def before_execute(self, command_name, *args):
        hook_map = self._hooks.get(command_name)
        if hook_map is not None:
            for hook in hook_map.get("before"):
                log.debug("Running before hook")
                hook_map["before"][hook].execute("before", command_name, args)

    def after_execute(self, command_name, *args):
        hook_map = self._hooks.get(command_name)
        if hook_map is not None:
            for hook in hook_map.get("after"):
                log.debug("Running after hook")
                hook_map["after"][hook].execute("after", command_name, args)

    def execute(self, command_name, *args, **kwargs):
        if command_name in self._commands.keys():
            log.debug(f"[EXCUTE] {command_name}")
            if self._history_position != 0 and self._unsaved_changes == False:
                self._get_sylk_json()
            # else:
            # self._sylk_json = self._history[self._history_position][0]
            self._history_position += 1
            self.before_execute(command_name, args)
            self._commands[command_name].execute(self._sylk_json, args, kwargs)
            self.after_execute(command_name, args)
            self._unsaved_changes = True
            if len(self._history) == self._history_position:
                # This is a new event in hisory
                temp = self._sylk_json.copy()
                self._history.append((temp, command_name, args, kwargs))
            else:
                # This occurs if there was one of more UNDOs and then a new
                # execute command happened. In case of UNDO, the history_position
                # changes, and executing new commands purges any history after
                # the current position"""
                self._history = self._history[: self._history_position + 1]
                self._history[self._history_position] = {
                    time.time(): [command_name, args]
                }
        else:
            log.error(f"Command [{command_name}] not recognised")

    def undo(self):
        """Undo a command if there is a command that can be undone.
        Update the history psoition so that further UNDOs or REDOs
        point to the correct index"""
        if self._history_position > 0:
            self._history_position -= 1
            log.debug(f"[UNDO] {self._history_position} / {len(self.history)}")
            self._sylk_json = self.history[self._history_position][0]
            # self._commands[
            #     self._history[self._history_position][1]
            # ].execute(self._history[self._history_position][0],self._history[self._history_position][2])
        else:
            log.warning("nothing to undo")

    def redo(self):
        """Perform a REDO if the history_position is less than the end of the history list"""
        if self._history_position + 1 < len(self._history):
            self._history_position += 1
            log.debug(f"[REDO] {self._history_position} / {len(self.history)}")
            self._commands[self._history[self._history_position][1]].execute(
                self._history[self._history_position][0],
                self._history[self._history_position][2],
            )
        else:
            log.warning("nothing to REDO")

    def save(self, sylkJson=True):
        """Saving current state"""
        current_state = self._sylk_json
        log.debug(f"[SAVE] {current_state}")
        self._unsaved_changes = False
        self._core_invoker.execute(
            "SaveSylkJson",
            current_state,
            {"format": self.sylkJson.get("configs").get("format", "json")},
        )
        self._saves += 1
        message = SylkJsonToMessage(self._sylk_json)
        path = join_path(
            self._sylk_json.get("project").get("uri"),
            ".sylk",
            "cache",
            f"save_{self._saves}.json",
        )
        if self._saves > 1:
            old_save = SylkJsonToMessage(
                rFile(
                    join_path(
                        self._sylk_json.get("project").get("uri"),
                        ".sylk",
                        "cache",
                        f"save_{self._saves-1}.json",
                    ),
                    True,
                )
            )
            if old_save != message:
                message = SylkJsonToMessage(self._sylk_json, validate=True)
                try:
                    wFile(path, MessageToDict(message), overwrite=True, json=True)
                except Exception:
                    temp_path = "/".join(path.split("/")[:-1])
                    mkdir(temp_path)
                    wFile(path, MessageToDict(message), overwrite=True, json=True)
                log.info(
                    f"Saving cache to [{join_path('.sylk','cache',f'save_{self._saves}.json')}]"
                )
        else:
            try:
                temp_p = path.replace("/cache/save_1.json", "")
                mkdir(temp_p)
                wFile(path, MessageToDict(message), overwrite=True, json=True)
            except Exception:
                temp_path = "/".join(path.split("/")[:-1])
                mkdir(temp_path)
                wFile(path, MessageToDict(message), overwrite=True, json=True)
            log.info(
                f"Saving cache to [{join_path('.sylk','cache',f'save_{self._saves}.json')}]"
            )

        # self._cache.insert(f'save_{self._saves}.json',MessageToDict(message))
        # out_s = open(join_path(self._sylk_json.get('project').get(
        #     'uri'), '.sylk', 'cache','dump'), 'wb')
        # try:
        #     log.debug('WRITING: %s' % (message.domain))
        #     pickle.dump(message, out_s)
        # finally:
        #     out_s.close()

    @property
    def history(self):
        return self._history

    @property
    def commands(self):
        return self._commands

    @property
    def hooks(self):
        return self._hooks

    @property
    def sylkJson(self):
        return self._sylk_json

    def validate_sylk_json(self, sylkJson):
        if sylkJson.packages is not None:
            for p in sylkJson.packages:
                pkg = sylkJson.packages[p]
                reorder = []
                index = 0
                for m in pkg.messages:
                    dependency_in_pkg = next(
                        (f for f in m.fields if f.message_type is not None), None
                    )

                    reorder.append(index)
                    if dependency_in_pkg is not None:
                        if pkg.package in dependency_in_pkg.message_type:
                            # pretty.print_error(dependency_in_pkg)
                            pretty.print_info(
                                reorder, True, "{0} / {1}".format(index, max(reorder))
                            )
                            if index > max(reorder):
                                reorder = [x + 1 for x in reorder]
                                reorder[index] = max(reorder) - 1
                            else:
                                reorder[index] = index + 1
                        else:
                            if index >= reorder[index]:
                                reorder = [x + 1 for x in reorder]

                            reorder[index] = index
                    else:
                        reorder[index] = max(reorder) + 1
                    pretty.print_info(dependency_in_pkg, True)
                    pretty.print_info(reorder, True, "After changes")

                    index += 1
                reorder = [x - 1 for x in reorder]
                mylist = [pkg.messages[i] for i in reorder]
                pkg_temp = MessageToDict(pkg)
                pkg_temp["messages"] = mylist

                pretty.print_info(mylist, True, "Last step")
                sylkJson_temp = MessageToDict(sylkJson)
                index = 0
                for i in pkg_temp["messages"]:
                    pretty.print_info(i, True)

                    pkg_temp[index] = MessageToDict(i)
                    index += 1

                sylkJson_temp.get("packages")[p] = pkg_temp
                pretty.print_info(sylkJson_temp, True)
                return helpers.SylkJsonToMessage(sylkJson_temp)
                # pretty.print_info(sylkJson.packages[p],True,"Package After Change step")
